// Copyright (c) 2013 by István Váradi

// This file is part of VSCPL, a simple cross-platform utility library

// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:

// 1. Redistributions of source code must retain the above copyright notice, this
//    list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright notice,
//    this list of conditions and the following disclaimer in the documentation
//    and/or other materials provided with the distribution.

// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
// ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// The views and conclusions contained in the software and documentation are those
// of the authors and should not be interpreted as representing official policies,
// either expressed or implied, of the FreeBSD Project.

#ifndef HU_VARADIISTVAN_SCPL_IO_POSIX_WAITER_H
#define HU_VARADIISTVAN_SCPL_IO_POSIX_WAITER_H
//------------------------------------------------------------------------------

#include "../Failable.h"

#include <map>

#include <cstdlib>

#include <poll.h>

//------------------------------------------------------------------------------

namespace hu { namespace varadiistvan { namespace scpl { namespace io {

//------------------------------------------------------------------------------

class Waitable;

//------------------------------------------------------------------------------

/**
 * Class that can be used to register waitable objects and then wait
 * for them.
 */
class Waiter : public Failable
{
private:
    /**
     * Type for the set of waitables.
     */
    typedef std::map<int, Waitable*> waitables_t;

    /**
     * The set of waitables belonging to this waiter.
     */
    waitables_t waitables;

    /**
     * The number of waitables.
     */
    size_t numWaitables;

public:
    /**
     * Construct the waiter.
     */
    Waiter();

    /**
     * Destroy the waiter. All waitables that still exists will be
     * modified to point to no waiter.
     */
    ~Waiter();

    /**
     * Wait for any if the waitables to become ready or until a
     * timeout expires.
     *
     * If any of the waitables are ready, the function returns immediately.
     *
     * @return true if any of the waitables has become ready, false if
     * the timeout occurs.
     */
    bool wait(int timeout = -1);

protected:
    /**
     * Add a waitable.
     */
    void add(Waitable* waitable);

    /**
     * Remove a waitable.
     */
    void remove(Waitable* waitable);

private:
    /**
     * Determine if any of the waitables is ready.
     */
    bool hasReady() const;

    /**
     * Fill the given array of poll descriptors with those file
     * descriptors that wait for an event.
     *
     * @param pollFDs the file descriptor array to fill. It should
     * contain at least as many elements as the number of waitables.
     *
     * @return the number of valid poll file descriptors.
     */
    size_t setupPollFDs(pollfd* pollFDs);

    /**
     * Process the given array of poll descriptors. For each
     * descriptor, if its events are not 0, the handleEvents()
     * function of the corresponding waitable is called.
     */
    void processPollFDs(const pollfd* pollFDs, size_t size);

    friend class Waitable;
};

//------------------------------------------------------------------------------
// Inline definitions
//------------------------------------------------------------------------------

inline Waiter::Waiter() :
    numWaitables(0)
{
}

//------------------------------------------------------------------------------

} /* namespace hu::varadiistvan::scpl::io */ } /* namespace hu::varadiistvan::scpl */ } /* namespace hu::varadiistvan */ } /* namespace hu */

//------------------------------------------------------------------------------
#endif // HU_VARADIISTVAN_SCPL_IO_POSIX_WAITER_H

// Local Variables:
// mode: C++
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
